@@ -36,7 +36,6 @@
 #include <unistd.h>
 #include <netdb.h>
 #include <net/if.h>
-#include <ifaddrs.h>
 #include <arpa/inet.h>
 
 #include <dbus/dbus-glib.h>
@@ -47,6 +46,12 @@
 #include "vino-http.h"
 #endif
 
+#ifdef HAVE_IFADDRS_H
+#include <ifaddrs.h>
+#else
+#include "libvncserver/ifaddr/ifaddrs.h"
+#endif
+
 #define VINO_DBUS_INTERFACE "org.gnome.VinoScreen"
 #define VINO_DBUS_BUS_NAME  "org.gnome.Vino"
 
