@@ -21,12 +21,14 @@
  */
 
 #include <config.h>
-
-#include "vino-prefs.h"
-
 #include <string.h>
 #include <gconf/gconf-client.h>
+#include <glib/gstdio.h>
 #include <signal.h>
+#include <fcntl.h>
+#include <stdlib.h>
+
+#include "vino-prefs.h"
 #include "vino-util.h"
 #include "vino-mdns.h"
 #include "vino-status-icon.h"
@@ -37,6 +39,7 @@
 #define VINO_PREFS_PROMPT_ENABLED         VINO_PREFS_DIR "/prompt_enabled"
 #define VINO_PREFS_VIEW_ONLY              VINO_PREFS_DIR "/view_only"
 #define VINO_PREFS_LOCAL_ONLY             VINO_PREFS_DIR "/local_only"
+#define VINO_PREFS_NETWORK_INTERFACE      VINO_PREFS_DIR "/network_interface"
 #define VINO_PREFS_USE_ALTERNATIVE_PORT   VINO_PREFS_DIR "/use_alternative_port"
 #define VINO_PREFS_ALTERNATIVE_PORT       VINO_PREFS_DIR "/alternative_port"
 #define VINO_PREFS_REQUIRE_ENCRYPTION     VINO_PREFS_DIR "/require_encryption"
@@ -45,8 +48,11 @@
 #define VINO_PREFS_LOCK_SCREEN            VINO_PREFS_DIR "/lock_screen_on_disconnect"
 #define VINO_PREFS_ICON_VISIBILITY        VINO_PREFS_DIR "/icon_visibility"
 #define VINO_PREFS_DISABLE_BACKGROUND     VINO_PREFS_DIR "/disable_background"
+#define VINO_PREFS_USE_UPNP               VINO_PREFS_DIR "/use_upnp"
 
-#define VINO_N_LISTENERS 12
+#define VINO_N_LISTENERS                  13
+
+#define VINO_PREFS_LOCKFILE               "vino-server.lock"
 
 static GConfClient *vino_client  = NULL;
 static GSList      *vino_servers = NULL;
@@ -55,7 +61,7 @@ static guint        vino_listeners [VINO_N_LISTENERS] = { 0, };
 static gboolean        vino_enabled              = FALSE;
 static gboolean        vino_prompt_enabled       = FALSE;
 static gboolean        vino_view_only            = FALSE;
-static gboolean        vino_local_only           = FALSE;
+static char           *vino_network_interface    = NULL;
 static gboolean        vino_require_encryption   = FALSE;
 static VinoAuthMethod  vino_auth_methods         = VINO_AUTH_VNC;
 static char           *vino_vnc_password         = NULL;
@@ -63,6 +69,7 @@ static gboolean        vino_use_alternative_port = FALSE;
 static int             vino_alternative_port     = VINO_SERVER_DEFAULT_PORT;
 static gboolean        vino_lock_screen          = FALSE;
 static gboolean        vino_disable_background   = FALSE;
+static gboolean        vino_use_upnp             = TRUE;
 static VinoStatusIconVisibility vino_icon_visibility = VINO_STATUS_ICON_VISIBILITY_CLIENT;
 
 static void
@@ -143,27 +150,35 @@ vino_prefs_view_only_changed (GConfClient *client,
 }
 
 static void
-vino_prefs_local_only_changed (GConfClient *client,
-			       guint	    cnxn_id,
-			       GConfEntry  *entry)
+vino_prefs_network_interface_changed (GConfClient *client,
+                                      guint        cnxn_id,
+                                      GConfEntry  *entry)
 {
-  gboolean  local_only;
-  GSList   *l;
+  const char *network_interface;
+  GSList     *l;
 
-  if (!entry->value || entry->value->type != GCONF_VALUE_BOOL)
+  if (!entry->value || entry->value->type != GCONF_VALUE_STRING)
     return;
 
-  local_only = gconf_value_get_bool (entry->value) != FALSE;
+  network_interface = gconf_value_get_string (entry->value);
 
-  if (vino_local_only == local_only)
+  if (!network_interface && !vino_network_interface)
     return;
 
-  vino_local_only = local_only;
+  if (network_interface && vino_network_interface &&
+      !strcmp (network_interface, vino_network_interface))
+    return;
 
-  dprintf (PREFS, "Local only changed: %s\n", vino_local_only ? "(true)" : "(false)");
+  if (vino_network_interface)
+    g_free (vino_network_interface);
+
+  vino_network_interface = g_strdup (network_interface);
+
+  dprintf (PREFS, "Network Interface for bind() changed: %s\n",
+       vino_network_interface ? vino_network_interface : "(null)");
 
   for (l = vino_servers; l; l = l->next)
-    vino_server_set_local_only (l->data, local_only);
+    vino_server_set_network_interface (l->data, network_interface);
 }
 
 static void
@@ -428,6 +443,30 @@ vino_prefs_disable_background_changed (GConfClient *client,
     vino_server_set_disable_background (l->data, disable_background);
 }
 
+static void
+vino_prefs_use_upnp_changed (GConfClient *client,
+                             guint       cnxn_id,
+                             GConfEntry  *entry)
+{
+  gboolean  use_upnp;
+  GSList   *l;
+
+  if (!entry->value || entry->value->type != GCONF_VALUE_BOOL)
+    return;
+
+  use_upnp = gconf_value_get_bool (entry->value) != FALSE;
+
+  if (vino_use_upnp == use_upnp)
+    return;
+
+  vino_use_upnp = use_upnp;
+
+  dprintf (PREFS, "Use UPNP changed: %s\n", vino_use_upnp ? "(true)" : "(false)");
+
+  for (l = vino_servers; l; l = l->next)
+    vino_server_set_use_upnp (l->data, use_upnp);
+}
+
 void
 vino_prefs_create_server (GdkScreen *screen)
 {
@@ -438,7 +477,7 @@ vino_prefs_create_server (GdkScreen *screen)
   server = g_object_new (VINO_TYPE_SERVER,
 			 "prompt-enabled",       vino_prompt_enabled,
 			 "view-only",            vino_view_only,
-			 "local-only",           vino_local_only,
+			 "network-interface",    vino_network_interface,
 			 "use-alternative-port", vino_use_alternative_port,
 			 "alternative-port",     vino_alternative_port,
 			 "auth-methods",         vino_auth_methods,
@@ -446,21 +485,104 @@ vino_prefs_create_server (GdkScreen *screen)
 			 "vnc-password",         vino_vnc_password,
 			 "on-hold",              !vino_enabled,
 			 "screen",               screen,
-             "lock-screen",          vino_lock_screen,
-             "disable-background",   vino_disable_background,
+			 "lock-screen",          vino_lock_screen,
+			 "disable-background",   vino_disable_background,
+			 "use-upnp",             vino_use_upnp,
 			 NULL);
 
   vino_servers = g_slist_prepend (vino_servers, server);
   if (vino_enabled)
     vino_mdns_start ();
 
-  g_signal_connect (server, "notify::alternative-port", vino_prefs_restart_mdns, NULL);
-  g_signal_connect (server, "notify::use-alternative-port", vino_prefs_restart_mdns, NULL);
+  g_signal_connect (server, "notify::alternative-port", G_CALLBACK (vino_prefs_restart_mdns), NULL);
+  g_signal_connect (server, "notify::use-alternative-port", G_CALLBACK(vino_prefs_restart_mdns), NULL);
 
   icon = vino_server_get_status_icon (server);
   vino_status_icon_set_visibility (icon, vino_icon_visibility);
 }
 
+static void
+vino_prefs_restore_background (void)
+{
+  if (vino_background_get_status ())
+    vino_background_draw (TRUE);
+}
+
+static gchar *
+vino_prefs_lock_filename (void)
+{
+  gchar *dir;
+
+  dir = g_build_filename (g_get_user_data_dir (),
+			 "vino",
+			  NULL);
+  if (!g_file_test (dir, G_FILE_TEST_EXISTS))
+    g_mkdir_with_parents (dir, 0755);
+
+  g_free (dir);
+
+  return g_build_filename (g_get_user_data_dir (),
+			   "vino",
+			    VINO_PREFS_LOCKFILE,
+			    NULL);
+}
+
+static gboolean
+vino_prefs_lock (void)
+{
+  gchar    *lockfile;
+  gboolean  res;
+
+  res = FALSE;
+  lockfile = vino_prefs_lock_filename ();
+
+  if (g_file_test (lockfile, G_FILE_TEST_EXISTS))
+    {
+      dprintf (PREFS, "WARNING: The lock file (%s) already exists\n", lockfile);
+    }
+  else
+    {
+      g_creat (lockfile, 0644);
+      res = TRUE;
+    }
+
+  g_free (lockfile);
+  return res;
+}
+
+static gboolean
+vino_prefs_unlock (void)
+{
+  gchar    *lockfile;
+  gboolean  res;
+
+  res = FALSE;
+  lockfile = vino_prefs_lock_filename ();
+
+  if (!g_file_test (lockfile, G_FILE_TEST_EXISTS))
+    {
+      dprintf (PREFS, "WARNING: Lock file (%s) not found!\n", lockfile);
+    }
+  else
+    {
+      g_unlink (lockfile);
+      res = TRUE;
+    }
+
+  g_free (lockfile);
+  return res;
+}
+
+static void
+vino_prefs_sighandler (int sig)
+{
+  g_message (_("Received signal %d, exiting...\n"), sig);
+  vino_prefs_restore_background ();
+  vino_mdns_shutdown ();
+  vino_prefs_shutdown ();
+  exit (0);
+}
+
 void
 vino_prefs_init (gboolean view_only)
 {
@@ -468,9 +590,10 @@ vino_prefs_init (gboolean view_only)
   int i = 0;
   char *key_str;
   
-  signal (SIGQUIT, vino_background_handler); /* Ctrl+C */
-  signal (SIGTERM, vino_background_handler); /* kill -15 */
-  signal (SIGSEGV, vino_background_handler); /* Segmentation fault */
+  signal (SIGINT,  vino_prefs_sighandler); /* Ctrl+C */
+  signal (SIGQUIT, vino_prefs_sighandler);
+  signal (SIGTERM, vino_prefs_sighandler); /* kill -15 */
+  signal (SIGSEGV, vino_prefs_sighandler); /* Segmentation fault */
 
   vino_client = gconf_client_get_default ();
 
@@ -479,6 +602,9 @@ vino_prefs_init (gboolean view_only)
 			GCONF_CLIENT_PRELOAD_ONELEVEL,
 			NULL);
 
+  if(!vino_prefs_lock ())
+    vino_prefs_restore_background ();
+
   vino_enabled = gconf_client_get_bool (vino_client, VINO_PREFS_ENABLED, NULL);
   dprintf (PREFS, "Access enabled: %s\n", vino_enabled ? "(true)" : "(false)");
 
@@ -498,15 +624,23 @@ vino_prefs_init (gboolean view_only)
 					      NULL);
     }
   dprintf (PREFS, "View only: %s\n", vino_view_only ? "(true)" : "(false)");
-      
-  vino_local_only = gconf_client_get_bool (vino_client,
-					   VINO_PREFS_LOCAL_ONLY,
-					   NULL);
-  dprintf (PREFS, "Local only: %s\n", vino_local_only ? "(true)" : "(false)");
+ 
+  vino_network_interface = gconf_client_get_string (vino_client,
+						    VINO_PREFS_NETWORK_INTERFACE,
+						    NULL);
+  /* Check for old key, local_only, vino <= 2.24 */
+  if (!vino_network_interface && gconf_client_get_bool (vino_client, VINO_PREFS_LOCAL_ONLY, NULL))
+    {
+      gconf_client_set_string (vino_client, VINO_PREFS_NETWORK_INTERFACE, "lo", NULL);
+      vino_network_interface = g_strdup ("lo");
+    }
+  dprintf (PREFS, "Network interface: %s\n", 
+                vino_network_interface ? vino_network_interface : "all");
 
   vino_use_alternative_port = gconf_client_get_bool (vino_client,
                                                      VINO_PREFS_USE_ALTERNATIVE_PORT,
                                                      NULL);
+    
   dprintf (PREFS, "Use alternative port: %s\n",
            vino_use_alternative_port ? "(true)" : "(false)");
 
@@ -551,6 +685,11 @@ vino_prefs_init (gboolean view_only)
                                                    NULL);
   dprintf (PREFS, "Disable background: %s\n", vino_disable_background ? "(true)" : "(false)");
 
+  vino_use_upnp = gconf_client_get_bool (vino_client,
+                                         VINO_PREFS_USE_UPNP,
+                                         NULL);
+  dprintf (PREFS, "Use UPNP: %s\n", vino_use_upnp ? "(true)" : "(false)");
+
   key_str = gconf_client_get_string (vino_client,
                                      VINO_PREFS_ICON_VISIBILITY,
                                      NULL);
@@ -584,8 +723,8 @@ vino_prefs_init (gboolean view_only)
 
   vino_listeners [i] =
     gconf_client_notify_add (vino_client,
-			     VINO_PREFS_LOCAL_ONLY,
-			     (GConfClientNotifyFunc) vino_prefs_local_only_changed,
+			     VINO_PREFS_NETWORK_INTERFACE,
+			     (GConfClientNotifyFunc) vino_prefs_network_interface_changed,
 			     NULL, NULL, NULL);
   i++;
 
@@ -649,6 +788,14 @@ vino_prefs_init (gboolean view_only)
 
   i++;
 
+  vino_listeners [i] =
+    gconf_client_notify_add (vino_client,
+			     VINO_PREFS_USE_UPNP,
+			     (GConfClientNotifyFunc) vino_prefs_use_upnp_changed,
+			     NULL, NULL, NULL);
+
+  i++;
+
   g_assert (i == VINO_N_LISTENERS);
 }
 
@@ -666,7 +813,11 @@ vino_prefs_shutdown (void)
   if (vino_vnc_password)
     g_free (vino_vnc_password);
   vino_vnc_password = NULL;
-      
+
+  if (vino_network_interface)
+    g_free (vino_network_interface);
+  vino_network_interface = NULL;
+
   for (i = 0; i < VINO_N_LISTENERS; i++) {
     if (vino_listeners [i])
       gconf_client_notify_remove (vino_client, vino_listeners [i]);
@@ -675,5 +826,6 @@ vino_prefs_shutdown (void)
 
   g_object_unref (vino_client);
   vino_client = NULL;
-}
 
+  vino_prefs_unlock ();
+}
