@@ -37,6 +37,7 @@
 #include "vino-util.h"
 #include "vino-enums.h"
 #include "vino-background.h"
+#include "vino-upnp.h"
 #include <sys/poll.h>
 #include <dbus/dbus-glib.h>
 #include <gtk/gtk.h>
@@ -60,9 +61,10 @@ struct _VinoServerPrivate
   VinoPrompt       *prompt;
   VinoStatusIcon   *icon;
   VinoDBusListener *listener;
+  VinoUpnp         *upnp;
 
-  GIOChannel       *io_channel;
-  guint             io_watch;
+  GIOChannel       *io_channel[RFB_MAX_SOCKETLISTEN];
+  guint             io_watch[RFB_MAX_SOCKETLISTEN];
 
   GSList           *clients;
 
@@ -76,7 +78,7 @@ struct _VinoServerPrivate
   int               alternative_port;
 
   guint             on_hold : 1;
-  guint             local_only : 1;
+  char             *network_interface;
   guint             prompt_enabled : 1;
   guint             view_only : 1;
   guint             require_encryption : 1;
@@ -84,6 +86,7 @@ struct _VinoServerPrivate
   guint             use_alternative_port : 1;
   guint             lock_screen : 1;
   guint             disable_background : 1;
+  guint             use_upnp : 1;
 };
 
 struct _VinoClient
@@ -106,7 +109,7 @@ enum
   PROP_ON_HOLD,
   PROP_PROMPT_ENABLED,
   PROP_VIEW_ONLY,
-  PROP_LOCAL_ONLY,
+  PROP_NETWORK_INTERFACE,
   PROP_USE_ALTERNATIVE_PORT,
   PROP_ALTERNATIVE_PORT,
   PROP_REQUIRE_ENCRYPTION,
@@ -114,7 +117,8 @@ enum
   PROP_VNC_PASSWORD,
   PROP_PORT,
   PROP_LOCK_SCREEN,
-  PROP_DISABLE_BACKGROUND
+  PROP_DISABLE_BACKGROUND,
+  PROP_USE_UPNP
 };
 
 static enum rfbNewClientAction vino_server_auth_client (VinoServer *server,
@@ -128,13 +132,14 @@ static void vino_server_update_security_types (VinoServer *server);
 
 static gpointer parent_class;
 
-static void
-vino_server_lock_screen (VinoServer *server)
-{
 #define GNOME_SCREENSAVER_BUS_NAME  "org.gnome.ScreenSaver"
 #define GNOME_SCREENSAVER_INTERFACE "org.gnome.ScreenSaver"
 #define GNOME_SCREENSAVER_PATH      "/org/gnome/ScreenSaver"
 
+static void
+vino_server_lock_screen (VinoServer *server)
+{
+
   DBusGConnection *connection;
   GError          *error;
   DBusGProxy      *proxy;
@@ -163,11 +168,42 @@ vino_server_lock_screen (VinoServer *server)
 
   g_object_unref (proxy);
   dbus_g_connection_unref (connection);
+}
+
+static void
+vino_server_unlock_screen (void)
+{
+
+  DBusGConnection *connection;
+  GError          *error;
+  DBusGProxy      *proxy;
+
+  dprintf(DBUS, "Unlocking screen via gnome-screensaver\n");
+
+  error = NULL;
+  connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
+  if (!connection)
+    {
+      g_printerr (_("Failed to open connection to bus: %s\n"),
+                  error->message);
+      g_error_free (error);
+      return;
+    }
+
+  proxy = dbus_g_proxy_new_for_name (connection,
+                                     GNOME_SCREENSAVER_BUS_NAME,
+                                     GNOME_SCREENSAVER_PATH,
+                                     GNOME_SCREENSAVER_INTERFACE);
+
+  dbus_g_proxy_call_no_reply (proxy, "SimulateUserActivity", G_TYPE_INVALID);
+
+  g_object_unref (proxy);
+  dbus_g_connection_unref (connection);
+}
 
 #undef GNOME_SCREENSAVER_BUS_NAME
 #undef GNOME_SCREENSAVER_INTERFACE
 #undef GNOME_SCREENSAVER_PATH
-}
 
 void
 vino_server_set_disable_background (VinoServer *server,
@@ -193,12 +229,55 @@ vino_server_get_disable_background (VinoServer *server)
   return server->priv->disable_background;
 }
 
+static void
+vino_server_control_upnp (VinoServer *server)
+{
+  if (server->priv->use_upnp && !server->priv->on_hold)
+    {
+      if (!server->priv->upnp)
+	server->priv->upnp = vino_upnp_new ();
+      vino_upnp_add_port (server->priv->upnp, server->priv->rfb_screen->rfbPort);
+    }
+  else
+    if (server->priv->upnp)
+      {
+	g_object_unref (server->priv->upnp);
+	server->priv->upnp = NULL;
+      }
+}
+
+void
+vino_server_set_use_upnp (VinoServer *server,
+                          gboolean use_upnp)
+{
+  g_return_if_fail (VINO_IS_SERVER (server));
+
+  use_upnp = use_upnp != FALSE;
+
+  if (server->priv->use_upnp != use_upnp)
+    {
+      server->priv->use_upnp = use_upnp;
+      vino_server_control_upnp (server);
+      g_object_notify (G_OBJECT (server), "use-upnp");
+    }
+}
+
+gboolean
+vino_server_get_use_upnp (VinoServer *server)
+{
+  g_return_val_if_fail (VINO_IS_SERVER (server), FALSE);
+
+  return server->priv->use_upnp;
+}
+
 static void
 vino_server_client_accepted (VinoServer *server,
                              VinoClient *client)
 {
   vino_status_icon_add_client (server->priv->icon, client);
 
+  vino_server_unlock_screen ();
+
   if (vino_server_get_disable_background (server))
      vino_background_draw (FALSE);
 }
@@ -250,8 +329,8 @@ vino_server_handle_client_gone (rfbClientPtr rfb_client)
 	    g_source_remove (client->update_timeout);
 	  client->update_timeout = 0;
 
-          if (client->io_watch)
-            g_source_remove (client->io_watch);
+    if (client->io_watch)
+      g_source_remove (client->io_watch);
 	  client->io_watch = 0;
 
 	  g_io_channel_unref (client->io_channel);
@@ -471,7 +550,7 @@ vino_server_new_connection_pending (GIOChannel   *source,
 {
   g_return_val_if_fail (VINO_IS_SERVER (server), FALSE);
 
-  rfbProcessNewConnection (server->priv->rfb_screen);
+  rfbProcessNewConnection (server->priv->rfb_screen, g_io_channel_unix_get_fd(source));
 
   return TRUE;
 }
@@ -813,6 +892,9 @@ vino_server_clipboard_cb (GtkClipboard *cb,
   gchar *text, *out;
   size_t a, b;
 
+  if (!server->priv->clients)
+    return;
+
   if (server->priv->on_hold)
     return;
 
@@ -833,7 +915,6 @@ vino_server_clipboard_cb (GtkClipboard *cb,
     }
 
   g_free (text);
-
 }
 
 static void
@@ -843,6 +924,7 @@ vino_server_init_from_screen (VinoServer *server,
   rfbScreenInfoPtr  rfb_screen;
   char             *name;
   GtkClipboard     *cb;
+  int               i;
 
   g_return_if_fail (server->priv->screen == NULL);
   g_return_if_fail (screen != NULL);
@@ -884,7 +966,7 @@ vino_server_init_from_screen (VinoServer *server,
    *   5900-6000
    */
   rfb_screen->rfbDeferUpdateTime = 0;
-  rfb_screen->localOnly          = server->priv->local_only;
+  rfb_screen->netIface           = server->priv->network_interface;
   rfb_screen->autoPort           = TRUE;
   rfb_screen->rfbPort            = VINO_SERVER_DEFAULT_PORT;
   rfb_screen->rfbAlwaysShared    = TRUE;
@@ -917,15 +999,18 @@ vino_server_init_from_screen (VinoServer *server,
 
   vino_server_update_security_types (server);
 
-  dprintf (RFB, "Creating watch for listening socket %d - port %d\n",
-	   rfb_screen->rfbListenSock, rfb_screen->rfbPort);
+  dprintf (RFB, "Creating watch for listening socket [ ");
+  for (i=0; i < rfb_screen->rfbListenSockTotal; i++)
+    {
+      dprintf (RFB, "%d ", rfb_screen->rfbListenSock[i]);
 
-  server->priv->io_channel = g_io_channel_unix_new (rfb_screen->rfbListenSock);
-  server->priv->io_watch =
-    g_io_add_watch (server->priv->io_channel,
-		    G_IO_IN|G_IO_PRI,
-		    (GIOFunc) vino_server_new_connection_pending,
-		    server);
+      server->priv->io_channel[i] = g_io_channel_unix_new (rfb_screen->rfbListenSock[i]);
+      server->priv->io_watch[i]   = g_io_add_watch (server->priv->io_channel[i],
+                                                    G_IO_IN|G_IO_PRI,
+                                                   (GIOFunc) vino_server_new_connection_pending,
+                                                    server);
+   }
+   dprintf(RFB,"]- port %d\n", rfb_screen->rfbPort);
 
 #ifdef VINO_ENABLE_HTTP_SERVER
   server->priv->http = vino_http_get (rfb_screen->rfbPort);
@@ -943,13 +1028,15 @@ vino_server_init_from_screen (VinoServer *server,
                     server);
 
   server->priv->icon = vino_status_icon_new (server, server->priv->screen);
+  server->priv->upnp = NULL;
 }
 
 static void
 vino_server_finalize (GObject *object)
 {
   VinoServer *server = VINO_SERVER (object);
-
+  int i;
+  
 #ifdef VINO_ENABLE_HTTP_SERVER
   if (server->priv->http)
     {
@@ -960,13 +1047,16 @@ vino_server_finalize (GObject *object)
   server->priv->http = NULL;
 #endif /* VINO_ENABLE_HTTP_SERVER */
 
-  if (server->priv->io_watch)
-    g_source_remove (server->priv->io_watch);
-  server->priv->io_watch = 0;
+  for(i=0; i < server->priv->rfb_screen->rfbListenSockTotal; i++)
+    {
+      if (server->priv->io_watch[i])
+        g_source_remove (server->priv->io_watch[i]);
+      server->priv->io_watch[i] = 0;
 
-  if (server->priv->io_channel)
-    g_io_channel_unref (server->priv->io_channel);
-  server->priv->io_channel = NULL;
+      if (server->priv->io_channel[i])
+        g_io_channel_unref (server->priv->io_channel[i]);
+      server->priv->io_channel[i] = NULL;
+    }
   
   if (server->priv->rfb_screen)
     rfbScreenCleanup (server->priv->rfb_screen);
@@ -981,6 +1071,10 @@ vino_server_finalize (GObject *object)
     g_free (server->priv->vnc_password);
   server->priv->vnc_password = NULL;
 
+  if(server->priv->network_interface)
+    g_free (server->priv->network_interface);
+  server->priv->network_interface = NULL;
+
   if (server->priv->prompt)
     g_object_unref (server->priv->prompt);
   server->priv->prompt = NULL;
@@ -1000,6 +1094,10 @@ vino_server_finalize (GObject *object)
   if (server->priv->icon)
     g_object_unref (server->priv->icon);
   server->priv->icon = NULL;
+
+  if (server->priv->upnp)
+    g_object_unref (server->priv->upnp);
+  server->priv->upnp = NULL;
   
   g_free (server->priv);
   server->priv = NULL;
@@ -1030,8 +1128,8 @@ vino_server_set_property (GObject      *object,
     case PROP_VIEW_ONLY:
       vino_server_set_view_only (server, g_value_get_boolean (value));
       break;
-    case PROP_LOCAL_ONLY:
-      vino_server_set_local_only (server, g_value_get_boolean (value));
+    case PROP_NETWORK_INTERFACE:
+      vino_server_set_network_interface (server, g_value_get_string (value));
       break;
     case PROP_REQUIRE_ENCRYPTION:
       vino_server_set_require_encryption (server, g_value_get_boolean (value));
@@ -1054,6 +1152,9 @@ vino_server_set_property (GObject      *object,
     case PROP_DISABLE_BACKGROUND:
       vino_server_set_disable_background (server, g_value_get_boolean (value));
       break;
+    case PROP_USE_UPNP:
+      vino_server_set_use_upnp (server, g_value_get_boolean (value));
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -1082,8 +1183,8 @@ vino_server_get_property (GObject    *object,
     case PROP_VIEW_ONLY:
       g_value_set_boolean (value, server->priv->view_only);
       break;
-    case PROP_LOCAL_ONLY:
-      g_value_set_boolean (value, server->priv->local_only);
+    case PROP_NETWORK_INTERFACE:
+      g_value_set_string (value, server->priv->network_interface);
       break;
     case PROP_REQUIRE_ENCRYPTION:
       g_value_set_boolean (value, server->priv->require_encryption);
@@ -1109,6 +1210,9 @@ vino_server_get_property (GObject    *object,
     case PROP_DISABLE_BACKGROUND:
       g_value_set_boolean (value, server->priv->disable_background);
       break;
+    case PROP_USE_UPNP:
+      g_value_set_boolean (value, server->priv->use_upnp);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -1181,11 +1285,11 @@ vino_server_class_init (VinoServerClass *klass)
                                                          G_PARAM_STATIC_BLURB));
 
   g_object_class_install_property (gobject_class,
-				   PROP_LOCAL_ONLY,
-				   g_param_spec_boolean ("local-only",
-							 "Local Only",
-							 "Only allow local connections",
-							 FALSE,
+				   PROP_NETWORK_INTERFACE,
+				   g_param_spec_string ("network-interface",
+							 "Network Interface",
+							 "Network interface for accept connections",
+							 NULL,
                                                          G_PARAM_READWRITE   |
                                                          G_PARAM_CONSTRUCT   |
                                                          G_PARAM_STATIC_NAME |
@@ -1293,6 +1397,18 @@ vino_server_class_init (VinoServerClass *klass)
                                                          G_PARAM_STATIC_NICK |
                                                          G_PARAM_STATIC_BLURB));
 
+  g_object_class_install_property (gobject_class,
+				   PROP_USE_UPNP,
+				   g_param_spec_boolean ("use-upnp",
+							 "Use UPNP",
+							 "Whether to use UPNP",
+							 FALSE,
+                                                         G_PARAM_READWRITE   |
+                                                         G_PARAM_CONSTRUCT   |
+                                                         G_PARAM_STATIC_NAME |
+                                                         G_PARAM_STATIC_NICK |
+                                                         G_PARAM_STATIC_BLURB));
+
 }
 
 GType
@@ -1368,32 +1484,32 @@ vino_server_set_view_only (VinoServer *server,
     }
 }
 
-gboolean
-vino_server_get_local_only (VinoServer *server)
+G_CONST_RETURN char *
+vino_server_get_network_interface (VinoServer *server)
 {
-  g_return_val_if_fail (VINO_IS_SERVER (server), FALSE);
-
-  return server->priv->local_only;
+  g_return_val_if_fail (VINO_IS_SERVER (server), NULL);
+  
+  return server->priv->network_interface;
 }
 
 void
-vino_server_set_local_only (VinoServer *server,
-                            gboolean    local_only)
+vino_server_set_network_interface (VinoServer *server,
+                                   const char *network_interface)
 {
   g_return_if_fail (VINO_IS_SERVER (server));
 
-  local_only = local_only != FALSE;
+  if (server->priv->network_interface)
+    g_free (server->priv->network_interface);
 
-  if (server->priv->local_only != local_only)
-    {
-      server->priv->local_only = local_only;
+  if(network_interface != NULL && strlen (network_interface) > 0)
+    server->priv->network_interface = g_strdup (network_interface);
+  else
+    server->priv->network_interface = NULL;
 
-      if (server->priv->rfb_screen != NULL)
-        rfbSetLocalOnly (server->priv->rfb_screen,
-                         server->priv->local_only);
+  if (server->priv->rfb_screen != NULL)
+    rfbSetNetworkInterface (server->priv->rfb_screen, server->priv->network_interface);
 
-      g_object_notify (G_OBJECT (server), "local-only");
-    }
+  g_object_notify (G_OBJECT (server), "network-interface");
 }
 
 gboolean
@@ -1424,9 +1540,12 @@ vino_server_set_use_alternative_port (VinoServer *server,
 
           rfbSetAutoPort (server->priv->rfb_screen,
                           !server->priv->use_alternative_port);
+
+	  vino_server_control_upnp (server);
         }
 
       g_object_notify (G_OBJECT (server), "use-alternative-port");
+      g_object_notify (G_OBJECT (server), "alternative-port");
     }
 }
 
@@ -1450,7 +1569,10 @@ vino_server_set_alternative_port (VinoServer *server,
 
       if (server->priv->rfb_screen &&
           server->priv->use_alternative_port)
-        rfbSetPort (server->priv->rfb_screen, server->priv->alternative_port);
+	{
+	  rfbSetPort (server->priv->rfb_screen, server->priv->alternative_port);
+	  vino_server_control_upnp (server);
+	}
 
       g_object_notify (G_OBJECT (server), "alternative-port");
     }
@@ -1495,9 +1617,11 @@ vino_server_set_on_hold (VinoServer *server,
 	}
 
       g_object_notify (G_OBJECT (server), "on-hold");
-      
+
       if (server->priv->icon)
         vino_status_icon_update_state (server->priv->icon);
+
+      vino_server_control_upnp (server);
     }
 }
 
@@ -1650,6 +1774,16 @@ vino_server_get_port (VinoServer *server)
   return server->priv->rfb_screen->rfbPort;
 }
 
+int
+vino_server_get_external_port (VinoServer *server)
+{
+  g_return_val_if_fail (VINO_IS_SERVER (server), 0);
+
+  return server->priv->use_upnp && VINO_IS_UPNP (server->priv->upnp) ?
+           vino_upnp_get_external_port (server->priv->upnp) :
+           server->priv->rfb_screen->rfbPort;
+}
+
 gboolean
 vino_server_get_lock_screen (VinoServer *server)
 {
@@ -1674,7 +1808,6 @@ vino_server_set_lock_screen (VinoServer *server,
     }
 }
 
-
 VinoStatusIcon *
 vino_server_get_status_icon (VinoServer *server)
 {
